@@ -4,8 +4,6 @@ Implements the MadVR protocol
 
 import logging
 from typing import Final, Union
-import re
-import time
 import asyncio
 from madvr.commands import ACKs, Footer, Commands, Enum, Connections
 from madvr.errors import AckError, RetryExceededError, HeartBeatError
@@ -28,35 +26,16 @@ class Madvr:
         self.connect_timeout: int = connect_timeout
         self.logger = logger
 
+        # used to indicate if connection is ready
+        self.connection_event = asyncio.Event()
+        self.stop_reconnect = asyncio.Event()
+
         # Const values
         self.MADVR_OK: Final = Connections.welcome.value
         self.HEARTBEAT: Final = Connections.heartbeat.value
 
-        # Incoming attrs
-        self.incoming_res = ""
-        self.incoming_frame_rate = ""
-        self.incoming_color_space = ""
-        self.incoming_bit_depth = ""
-        self.hdr_flag = False
-        self.incoming_colorimetry = ""
-        self.incoming_black_levels = ""
-        self.incoming_aspect_ratio = ""
-        self.aspect_ratio: float = 0
-
-        # Temps
-        self.temp_gpu: int = 0
-        self.temp_hdmi: int = 0
-        self.temp_cpu: int = 0
-        self.temp_mainboard: int = 0
-
-        # Outgoing signal
-        self.outgoing_res = ""
-        self.outgoing_frame_rate = ""
-        self.outgoing_color_space = ""
-        self.outgoing_bit_depth = ""
-        self.outgoing_colorimetry = ""
-        self.outgoing_hdr_flag = False
-        self.outgoing_black_levels = ""
+        # stores all attributes
+        self.msg_dict = {}
 
         # Sockets
         self.reader = None
@@ -64,15 +43,12 @@ class Madvr:
         self.reader_lock = asyncio.Lock()
         self.writer_lock = asyncio.Lock()
 
-        self.notification_reader = None
-        self.notification_writer = None
-        self.notification_reader_lock = asyncio.Lock()
-        self.notification_writer_lock = asyncio.Lock()
         self.is_closed = False
         # Envy does not have an are you on cmd, just assuming its on based on active connection
         self.is_on = False
         self.read_limit = 8000
         self.command_read_timeout = 3
+
         self.logger.debug("Running in debug mode")
 
     async def _clear_attr(self) -> None:
@@ -80,30 +56,7 @@ class Madvr:
         Clear instance attr so HA doesn't report stale values
         """
         # Incoming attrs
-        self.incoming_res = ""
-        self.incoming_frame_rate = ""
-        self.incoming_color_space = ""
-        self.incoming_bit_depth = ""
-        self.hdr_flag = False
-        self.incoming_colorimetry = ""
-        self.incoming_black_levels = ""
-        self.incoming_aspect_ratio = ""
-        self.aspect_ratio = 0
-
-        # Temps
-        self.temp_gpu = 0
-        self.temp_hdmi = 0
-        self.temp_cpu = 0
-        self.temp_mainboard = 0
-
-        # Outgoing signal
-        self.outgoing_res = ""
-        self.outgoing_frame_rate = ""
-        self.outgoing_color_space = ""
-        self.outgoing_bit_depth = ""
-        self.outgoing_colorimetry = ""
-        self.outgoing_hdr_flag = False
-        self.outgoing_black_levels = ""
+        self.msg_dict = {}
 
     async def close_connection(self) -> None:
         """close the connection"""
@@ -114,18 +67,10 @@ class Madvr:
             except AttributeError:
                 # means its already closed
                 pass
-        async with self.notification_writer_lock:
-            try:
-                self.notification_writer.close()
-                await self.notification_writer.wait_closed()
-            except AttributeError:
-                # means its already closed
-                pass
 
         self.reader = None
-        self.notification_reader = None
-
         self.is_closed = True
+        self.connection_event.clear()
 
         # Clear attr
         await self._clear_attr()
@@ -138,30 +83,25 @@ class Madvr:
         except AckError as err:
             self.logger.error(err)
 
+    def stop(self):
+        """Stop reconnecting"""
+        self.stop_reconnect.set()
+
     async def _reconnect(self) -> None:
         """
         Initiate keep-alive connection. This should handle any error and reconnect eventually.
 
         Raises AckError
         """
-
-        while True:
+        while not self.stop_reconnect.is_set():
             try:
                 self.logger.info("Connecting to Envy: %s:%s", self.host, self.port)
 
                 # Command client
-                async with self.reader_lock, self.notification_reader_lock:
+                async with self.reader_lock:
                     self.reader, self.writer = await asyncio.wait_for(
                         asyncio.open_connection(self.host, self.port),
-                        timeout=self.command_read_timeout,
-                    )
-
-                    # Notifications client
-                    (
-                        self.notification_reader,
-                        self.notification_writer,
-                    ) = await asyncio.wait_for(
-                        asyncio.open_connection(self.host, self.port), timeout=20
+                        timeout=2,
                     )
 
                 # Test heartbeat
@@ -181,29 +121,20 @@ class Madvr:
                         f"Notification did not reply with correct greeting: {msg_envy}"
                     )
 
-                # Make sure first message says WELCOME for notifications
-                async with self.notification_reader_lock:
-                    msg_envy = await asyncio.wait_for(
-                        self.notification_reader.readline(), timeout=20
-                    )
-
-                # Check if first 7 char match
-                if self.MADVR_OK != msg_envy[:7]:
-                    raise AckError(
-                        f"Notification did not reply with correct greeting: {msg_envy}"
-                    )
                 self.logger.info("Waiting for envy to be available")
                 # envy needs some time to setup new connections
                 await asyncio.sleep(3)
 
                 # handshake func
-                await self._send_heartbeat()
+                await self.send_heartbeat(True)
 
                 self.logger.info("Connection established")
 
                 self.is_on = True
                 self.is_closed = False
 
+                self.connection_event.set()
+
                 return
 
             except HeartBeatError:
@@ -228,7 +159,7 @@ class Madvr:
                 await asyncio.sleep(2)
                 continue
 
-    async def _send_heartbeat(self) -> None:
+    async def send_heartbeat(self, once=False) -> None:
         """
         Send a heartbeat to keep connection open
 
@@ -236,40 +167,41 @@ class Madvr:
 
         Raises HeartBeatError exception
         """
-        i = 0
+        if once:
+            try:
+                async with self.writer_lock:
+                    self.writer.write(self.HEARTBEAT)
+                    await self.writer.drain()
 
-        while i < 3:
-            # confirm can send heartbeat, ready for commands
-            self.logger.debug("Sending heartbeats")
-
-            async with self.writer_lock:
-                self.writer.write(self.HEARTBEAT)
-                await self.writer.drain()
-
-            # Recv until we find ok or timeout
-            if not await self._read_until_ok(self.reader):
-                i += 1
-                continue
-
-            # send heartbeat with notification and read ack and then regular client
-            # ensure both work one after the next
-            async with self.notification_writer_lock:
-                self.notification_writer.write(self.HEARTBEAT)
-                await self.notification_writer.drain()
-
-            if not await self._read_until_ok(self.notification_reader):
-                i += 1
-                continue
-
-            self.logger.debug("Handshakes complete")
+                self.logger.debug("heartbeat complete")
+            except asyncio.TimeoutError:
+                self.logger.error("timeout when sending heartbeat")
+            except OSError:
+                self.logger.error("error when sending heartbeat")
+                await self._reconnect()
 
             return
 
-        raise HeartBeatError("Sending heartbeat fatal error")
+        while True:
+            await self.connection_event.wait()
+            # confirm can send heartbeat, ready for commands
+            self.logger.debug("Sending heartbeats")
+            try:
+                async with self.writer_lock:
+                    self.writer.write(self.HEARTBEAT)
+                    await self.writer.drain()
 
-    async def _construct_command(
-        self, raw_command: Union[str, list]
-    ) -> tuple[bytes, str]:
+                self.logger.debug("heartbeat complete")
+            except asyncio.TimeoutError:
+                self.logger.error("timeout when sending heartbeat")
+            except OSError:
+                self.logger.error("error when sending heartbeat")
+                await self._reconnect()
+            finally:
+                # Wait some time before the next heartbeat
+                await asyncio.sleep(15)
+
+    async def _construct_command(self, raw_command: list[str]) -> tuple[bytes, str]:
         """
         Transform commands into their byte values from the string value
 
@@ -287,7 +219,7 @@ class Madvr:
         # HA seems to always send commands as a list even if you set them as a str
 
         # This lets you use single cmds or something with val like KEYPRESS
-        # If len is 1, then try to split, otherwise its just one word
+        # If len is 1 like ["keypress,val"], then try to split, otherwise its just one word
 
         if len(raw_command) == 1:
             try:
@@ -300,7 +232,7 @@ class Madvr:
             except ValueError as err:
                 self.logger.debug(err)
                 self.logger.debug("Using raw_command directly")
-                command = raw_command
+                command = raw_command[0]
                 skip_val = True
         # if there are more than two values, this is incorrect, error
         elif len(raw_command) > 2:
@@ -313,12 +245,12 @@ class Madvr:
             # raw command will be a list of 2
             command, value = raw_command
 
-        self.logger.debug("using command %s", command)
+        self.logger.debug("checking command %s", command)
 
         # Check if command is implemented
         if not hasattr(Commands, command):
             raise NotImplementedError(f"Command not implemented: {command}")
-
+        self.logger.debug("Found command")
         # construct the command with nested Enums
         command_name, val, _ = Commands[command].value
 
@@ -339,43 +271,20 @@ class Madvr:
 
         return cmd, val
 
-    async def _read_until_ok(self, client: asyncio.StreamReader) -> bool:
-        """
-        Read the buffer until we get ok or timeout because a timeout means no more to read
-        """
-        counter = 0
-        # exit loop if exceed maximum checks for ok or retries
-        while True:
-            try:
-                async with self.reader_lock, self.notification_reader_lock:
-                    data = await asyncio.wait_for(client.readline(), timeout=10)
-                    self.logger.debug("_read_until_ok: data found: %s", data)
-
-                if ACKs.reply.value not in data:
-                    self.logger.debug("OK not found yet counter: %s", counter)
-                    counter += 1
-                    continue
-
-                self.logger.debug("OK found counter: %s", counter)
-                return True
-
-            except asyncio.TimeoutError:
-                self.logger.debug("Timeout reading ack with counter: %s", counter)
-                return False
-
-    async def send_command(self, command: Union[str, list]) -> str:
+    async def send_command(self, command: list) -> str:
         """
         send a given command same as the official madvr ones
         To keep this simple, just send the command without reading response
 
-        command: str - command to send like KeyPress, MENU
+        command: list - command to send like [KeyPress, MENU]
         Raises RetryExceededError
         """
 
         # Verify the command is supported
         try:
             cmd, enum_type = await self._construct_command(command)
-        except NotImplementedError:
+        except NotImplementedError as err:
+            self.logger.error("command not implemented: %s -- %s", command, err)
             return f"Command not found: {command}"
 
         self.logger.debug("using values: %s %s", cmd, enum_type)
@@ -392,151 +301,92 @@ class Madvr:
         retry_count = 0
 
         while retry_count < 5:
-            # Send the command
             try:
                 async with self.writer_lock:
                     self.writer.write(cmd)
                     await self.writer.drain()
-
-            except asyncio.TimeoutError:
-                self.logger.debug("OK receipt timed out, retrying")
+                break  # if success, break the loop
+            except (asyncio.TimeoutError, OSError):
+                self.logger.debug(
+                    "OK receipt timed out or connection failed when reading OK, retrying"
+                )
                 retry_count += 1
+                await asyncio.sleep(1)  # sleep before retrying
                 continue
 
-            # should catch connection is closed
-            except OSError:
-                self.logger.warning("connection failed when reading OK, retrying")
-                retry_count += 1
-                continue
-
-            return ""
-
-        # if we got here something went wrong
-        await self.close_connection()
-        await self._reconnect()
+        if retry_count == 5:  # if we got here something went wrong
+            await self.close_connection()
+            await self._reconnect()
 
         return ""
 
-    async def start_read_notifications(self, wait_forever: bool) -> None:
+    async def read_notifications(self) -> None:
         """
-        Listen for notifications. Meant to run as a background task
-        wait_forever: bool -> if true, it will block forever. False useful for testing
+        Read notifications from the server and update attributes
         """
-        # reconnect if client is not init or its off
-        if self.notification_reader is None or self.is_on is False:
-            await self._reconnect()
-
-        # Receive data in a loop
-        i = 0
-        while wait_forever or i < 5:
+        while True:
+            # wait until the connection is established
+            await self.connection_event.wait()
             try:
-                # send hearbeat
-                async with self.notification_writer_lock:
-                    self.notification_writer.write(self.HEARTBEAT)
-                    await self.notification_writer.drain()
-                
-                async with self.notification_reader_lock:
-                    data = await asyncio.wait_for(
-                        self.reader.readline(),
+                async with self.reader_lock:
+                    msg = await asyncio.wait_for(
+                        self.reader.read(self.read_limit),
                         timeout=self.command_read_timeout,
                     )
-
-                asyncio.sleep(1)
-
-                i += 1
-
-                if data:
-                    # process data which will get added as class attr
-                    await self._process_notifications(data)
-                else:
-                    # just wait forever for data
-                    # send heartbeat keep conn open
-                    async with self.notification_writer_lock:
-                        self.notification_writer.write(self.HEARTBEAT)
-                        await self.notification_writer.drain()
-                    continue
-            except asyncio.TimeoutError:
-                self.logger.debug("Connection timed out")
-                async with self.notification_writer_lock:
-                    self.notification_writer.write(self.HEARTBEAT)
-                    await self.notification_writer.drain()
-                continue
-            except OSError:
-                self.logger.debug("Connection error")
-                async with self.notification_writer_lock:
-                    self.notification_writer.write(self.HEARTBEAT)
-                    await self.notification_writer.drain()
-                continue
-            except AttributeError:
+            except ConnectionResetError:
+                self.logger.warning(
+                    "Connection reset by peer. Attempting to reconnect..."
+                )
                 await self._reconnect()
+            except (AttributeError, asyncio.TimeoutError, OSError) as err:
+                self.logger.warning(
+                    "Reading notifications failed or timed out: %s", err
+                )
+                await asyncio.sleep(5)
+                continue
 
-    async def _process_notifications(self, input_data: Union[bytes, str]) -> None:
-        """
-        Process arbitrary stream of notifications and set them as instance attr
-        """
-        # This code constructs a dict for all values processed
-        self.logger.debug("Processing data for %s", input_data)
-        try:
-            if isinstance(input_data, bytes):
-                # This pattern will be able to extract from the byte encoded stream
-                pattern = r"([A-Z][^\r\n]*)\r\n"
-                groups = re.findall(pattern, input_data.decode())
-                # split the groups, the first element is the key, remove the key from the values
-                # for each match in groups, add it to dict
-                # {"key": ["val1", "val2"]}
-                val_dict: dict = {
-                    group.split()[0]: group.replace(group.split()[0] + " ", "").split()
-                    for group in groups
-                }
-                self.logger.debug("groups: %s", groups)
-            else:
-                self.logger.debug("input data: %s", input_data)
-                # This pattern extracts from a regular string
-                # If we have a str its assumed we are dealing with one output stream
-                pattern = r"([A-Z][A-Za-z]*)\s(.*)"
-                match = re.search(pattern, input_data)
-                val_dict: dict = {match.group(1): match.group(2).split()}
+            if not msg:
+                await asyncio.sleep(1)
+                continue
 
-            self.logger.debug("val dict: %s", val_dict)
-        except AttributeError:
-            return
+            await self._process_notifications(msg.decode("utf-8"))
 
-        # Map values to attr
-        incoming_signal_info: list = val_dict.get("IncomingSignalInfo", [])
-        if incoming_signal_info:
-            self.logger.debug("incoming signal detected: %s", incoming_signal_info)
-            self.incoming_res = incoming_signal_info[0]
-            self.incoming_frame_rate = incoming_signal_info[1]
-            self.incoming_color_space = incoming_signal_info[3]
-            self.incoming_bit_depth = incoming_signal_info[4]
-            self.hdr_flag = "HDR" in incoming_signal_info[5]
-            self.incoming_colorimetry = incoming_signal_info[6]
-            self.incoming_black_levels = incoming_signal_info[7]
-            self.incoming_aspect_ratio = incoming_signal_info[8]
-
-        aspect_ratio: list = val_dict.get("AspectRatio", [])
-        if aspect_ratio:
-            self.logger.debug("incoming AR detected: %s", aspect_ratio)
-            self.aspect_ratio = float(aspect_ratio[1])
-
-        get_temps: list = val_dict.get("Temperatures", [])
-        if get_temps:
-            self.logger.debug("incoming Temps detected: %s", get_temps)
-            self.temp_gpu = int(get_temps[0])
-            self.temp_hdmi = int(get_temps[1])
-            self.temp_cpu = int(get_temps[2])
-            self.temp_mainboard = int(get_temps[3])
-
-        outgoing_signal_info: list = val_dict.get("OutgoingSignalInfo", [])
-        if outgoing_signal_info:
-            self.logger.debug("outgoing signal detected: %s", outgoing_signal_info)
-            self.outgoing_res = outgoing_signal_info[0]
-            self.outgoing_frame_rate = outgoing_signal_info[1]
-            self.outgoing_color_space = outgoing_signal_info[3]
-            self.outgoing_bit_depth = outgoing_signal_info[4]
-            self.outgoing_hdr_flag = "HDR" in outgoing_signal_info[5]
-            self.outgoing_colorimetry = outgoing_signal_info[6]
-            self.outgoing_black_levels = outgoing_signal_info[7]
+    async def _process_notifications(self, msg: str) -> None:
+        """Parse a message and store the attributes and values in a dictionary"""
+        self.logger.debug("Processing notifications: %s", msg)
+        notifications = msg.split("\r\n")
+        # for each /r/n split it by title, then the rest are values
+        for notification in notifications:
+            title, *signal_info = notification.split(" ")
+            # dont process empty values
+            if not signal_info:
+                continue
+            # at least madvr sends attributes in a consistent order
+            # could use zip here but why? this works and is simple
+            if "NoSignal" in title:
+                return
+            if "IncomingSignalInfo" in title:
+                self.msg_dict["incoming_res"] = signal_info[0]
+                self.msg_dict["incoming_frame_rate"] = signal_info[1]
+                self.msg_dict["incoming_color_space"] = signal_info[3]
+                self.msg_dict["incoming_bit_depth"] = signal_info[4]
+                self.msg_dict["hdr_flag"] = "HDR" in signal_info[5]
+                self.msg_dict["incoming_colorimetry"] = signal_info[6]
+                self.msg_dict["incoming_black_levels"] = signal_info[7]
+                self.msg_dict["incoming_aspect_ratio"] = signal_info[8]
+            elif "OutgoingSignalInfo" in title:
+                self.msg_dict["outgoing_res"] = signal_info[0]
+                self.msg_dict["outgoing_frame_rate"] = signal_info[1]
+                self.msg_dict["outgoing_color_space"] = signal_info[3]
+                self.msg_dict["outgoing_bit_depth"] = signal_info[4]
+                self.msg_dict["outgoing_hdr_flag"] = "HDR" in signal_info[5]
+                self.msg_dict["outgoing_colorimetry"] = signal_info[6]
+                self.msg_dict["outgoing_black_levels"] = signal_info[7]
+            elif "AspectRatio" in title:
+                self.msg_dict["aspect_res"] = signal_info[0]
+                self.msg_dict["aspect_dec"] = float(signal_info[1])
+                self.msg_dict["aspect_int"] = signal_info[2]
+                self.msg_dict["aspect_name"] = signal_info[3]
 
     async def power_off(self, standby=False) -> str:
         """
